@@ -1,6 +1,6 @@
 /* Save the list of files in a directory in a string.
 
-   Copyright 1997, 1999, 2001, 2003 Free Software Foundation, Inc.
+   Copyright 1997, 1999, 2001, 2003, 2005 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -14,13 +14,14 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by David MacKenzie <djm@gnu.ai.mit.edu>. */
 
 #if !defined SAVEDIR_H_
 # define SAVEDIR_H_
 
-char *savedir (const char *dir);
+char *savedir (char const *dir);
+char *fdsavedir (int fd);
 
 #endif
