@@ -1,5 +1,5 @@
-/* Compute MD5 or SHA1 checksum of files or strings
-   Copyright (C) 1995-2004 Free Software Foundation, Inc.
+/* Compute MD5, SHA1, SHA224, SHA256, SHA384 or SHA512 checksum of files or strings
+   Copyright (C) 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,68 +13,90 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 /* Written by Ulrich Drepper <drepper@gnu.ai.mit.edu>.  */
 
 #include <config.h>
 
 #include <getopt.h>
-#include <stdio.h>
 #include <sys/types.h>
 
 #include "system.h"
 
-#include "md5.h"
-#include "sha1.h"
-#include "checksum.h"
+#if HASH_ALGO_MD5
+# include "md5.h"
+#endif
+#if HASH_ALGO_SHA1
+# include "sha1.h"
+#endif
+#if HASH_ALGO_SHA256 || HASH_ALGO_SHA224
+# include "sha256.h"
+#endif
+#if HASH_ALGO_SHA512 || HASH_ALGO_SHA384
+# include "sha512.h"
+#endif
 #include "getline.h"
 #include "error.h"
 #include "quote.h"
+#include "stdio--.h"
 
 /* The official name of this program (e.g., no `g' prefix).  */
-#define PROGRAM_NAME (algorithm == ALG_MD5 ? "md5sum" : "sha1sum")
-
-#define AUTHORS "Ulrich Drepper", "Scott Miller"
-
-/* Most systems do not distinguish between external and internal
-   text representations.  */
-/* FIXME: This begs for an autoconf test.  */
-#if O_BINARY
-# define OPENOPTS(BINARY) ((BINARY) ? TEXT1TO1 : TEXTCNVT)
-# define TEXT1TO1 "rb"
-# define TEXTCNVT "r"
+#if HASH_ALGO_MD5
+# define PROGRAM_NAME "md5sum"
+# define DIGEST_TYPE_STRING "MD5"
+# define DIGEST_STREAM md5_stream
+# define DIGEST_BITS 128
+# define DIGEST_REFERENCE "RFC 1321"
+# define DIGEST_ALIGN 4
+#elif HASH_ALGO_SHA1
+# define PROGRAM_NAME "sha1sum"
+# define DIGEST_TYPE_STRING "SHA1"
+# define DIGEST_STREAM sha1_stream
+# define DIGEST_BITS 160
+# define DIGEST_REFERENCE "FIPS-180-1"
+# define DIGEST_ALIGN 4
+#elif HASH_ALGO_SHA256
+# define PROGRAM_NAME "sha256sum"
+# define DIGEST_TYPE_STRING "SHA256"
+# define DIGEST_STREAM sha256_stream
+# define DIGEST_BITS 256
+# define DIGEST_REFERENCE "FIPS-180-2"
+# define DIGEST_ALIGN 4
+#elif HASH_ALGO_SHA224
+# define PROGRAM_NAME "sha224sum"
+# define DIGEST_TYPE_STRING "SHA224"
+# define DIGEST_STREAM sha224_stream
+# define DIGEST_BITS 224
+# define DIGEST_REFERENCE "RFC 3874"
+# define DIGEST_ALIGN 4
+#elif HASH_ALGO_SHA512
+# define PROGRAM_NAME "sha512sum"
+# define DIGEST_TYPE_STRING "SHA512"
+# define DIGEST_STREAM sha512_stream
+# define DIGEST_BITS 512
+# define DIGEST_REFERENCE "FIPS-180-2"
+# define DIGEST_ALIGN 8
+#elif HASH_ALGO_SHA384
+# define PROGRAM_NAME "sha384sum"
+# define DIGEST_TYPE_STRING "SHA384"
+# define DIGEST_STREAM sha384_stream
+# define DIGEST_BITS 384
+# define DIGEST_REFERENCE "FIPS-180-2"
+# define DIGEST_ALIGN 8
 #else
-# if defined VMS
-#  define OPENOPTS(BINARY) ((BINARY) ? TEXT1TO1 : TEXTCNVT)
-#  define TEXT1TO1 "rb", "ctx=stm"
-#  define TEXTCNVT "r", "ctx=stm"
-# else
-#  if UNIX || __UNIX__ || unix || __unix__ || _POSIX_VERSION
-#   define OPENOPTS(BINARY) "r"
-#  else
-    /* The following line is intended to evoke an error.
-       Using #error is not portable enough.  */
-    "Cannot determine system type."
-#  endif
-# endif
+# error "Can't decide which hash algorithm to compile."
 #endif
 
+#define DIGEST_HEX_BYTES (DIGEST_BITS / 4)
+#define DIGEST_BIN_BYTES (DIGEST_BITS / 8)
 
-#define DIGEST_TYPE_STRING(Alg) ((Alg) == ALG_MD5 ? "MD5" : "SHA1")
-#define DIGEST_STREAM(Alg) ((Alg) == ALG_MD5 ? md5_stream : sha1_stream)
-
-#define DIGEST_BITS(Alg) ((Alg) == ALG_MD5 ? 128 : 160)
-#define DIGEST_HEX_BYTES(Alg) (DIGEST_BITS (Alg) / 4)
-#define DIGEST_BIN_BYTES(Alg) (DIGEST_BITS (Alg) / 8)
-
-#define MAX_DIGEST_BIN_BYTES MAX (DIGEST_BIN_BYTES (ALG_MD5), \
-				  DIGEST_BIN_BYTES (ALG_SHA1))
+#define AUTHORS "Ulrich Drepper", "Scott Miller", "David Madore"
 
 /* The minimum length of a valid digest line.  This length does
    not include any newline character at the end of a line.  */
-#define MIN_DIGEST_LINE_LENGTH(Alg) \
-  (DIGEST_HEX_BYTES (Alg) /* length of hexadecimal message digest */ \
+#define MIN_DIGEST_LINE_LENGTH \
+  (DIGEST_HEX_BYTES /* length of hexadecimal message digest */ \
    + 2 /* blank and binary indicator */ \
    + 1 /* minimum filename length */ )
 
@@ -95,9 +117,6 @@ static bool status_only = false;
    improperly formatted checksum line.  */
 static bool warn = false;
 
-/* Declared and set via one of the wrapper .c files.  */
-/* int algorithm = ALG_UNSPECIFIED; */
-
 /* The name this program was run with.  */
 char *program_name;
 
@@ -110,11 +129,11 @@ enum
 
 static const struct option long_options[] =
 {
-  { "binary", no_argument, 0, 'b' },
-  { "check", no_argument, 0, 'c' },
-  { "status", no_argument, 0, STATUS_OPTION },
-  { "text", no_argument, 0, 't' },
-  { "warn", no_argument, 0, 'w' },
+  { "binary", no_argument, NULL, 'b' },
+  { "check", no_argument, NULL, 'c' },
+  { "status", no_argument, NULL, STATUS_OPTION },
+  { "text", no_argument, NULL, 't' },
+  { "warn", no_argument, NULL, 'w' },
   { GETOPT_HELP_OPTION_DECL },
   { GETOPT_VERSION_OPTION_DECL },
   { NULL, 0, NULL, 0 }
@@ -130,25 +149,37 @@ usage (int status)
     {
       printf (_("\
 Usage: %s [OPTION] [FILE]...\n\
-  or:  %s [OPTION] --check [FILE]\n\
 Print or check %s (%d-bit) checksums.\n\
 With no FILE, or when FILE is -, read standard input.\n\
+\n\
 "),
-	      program_name, program_name,
-	      DIGEST_TYPE_STRING (algorithm),
-	      DIGEST_BITS (algorithm));
+	      program_name,
+	      DIGEST_TYPE_STRING,
+	      DIGEST_BITS);
+      if (O_BINARY)
+	fputs (_("\
+  -b, --binary            read in binary mode (default unless reading tty stdin)\n\
+"), stdout);
+      else
+	fputs (_("\
+  -b, --binary            read in binary mode\n\
+"), stdout);
       printf (_("\
-\n\
-  -b, --binary            read files in binary mode (default on DOS/Windows)\n\
-  -c, --check             check %s sums against given list\n\
-  -t, --text              read files in text mode (default)\n\
-\n\
-"),
-	      DIGEST_TYPE_STRING (algorithm));
+  -c, --check             read %s sums from the FILEs and check them\n"),
+	      DIGEST_TYPE_STRING);
+      if (O_BINARY)
+	fputs (_("\
+  -t, --text              read in text mode (default if reading tty stdin)\n\
+"), stdout);
+      else
+	fputs (_("\
+  -t, --text              read in text mode (default)\n\
+"), stdout);
       fputs (_("\
+\n\
 The following two options are useful only when verifying checksums:\n\
       --status            don't output anything, status code shows success\n\
-  -w, --warn              warn about improperly formated checksum lines\n\
+  -w, --warn              warn about improperly formatted checksum lines\n\
 \n\
 "), stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
@@ -159,7 +190,7 @@ The sums are computed as described in %s.  When checking, the input\n\
 should be a former output of this program.  The default mode is to print\n\
 a line with checksum, a character indicating type (`*' for binary, ` ' for\n\
 text), and name for each FILE.\n"),
-	      (algorithm == ALG_MD5 ? "RFC 1321" : "FIPS-180-1"));
+	      DIGEST_REFERENCE);
       printf (_("\nReport bugs to <%s>.\n"), PACKAGE_BUGREPORT);
     }
 
@@ -211,7 +242,7 @@ bsd_split_3 (char *s, size_t s_len, unsigned char **hex_digest, char **file_name
 
 static bool
 split_3 (char *s, size_t s_len,
-	 unsigned char **hex_digest, bool *binary, char **file_name)
+	 unsigned char **hex_digest, int *binary, char **file_name)
 {
   size_t i;
   bool escaped_filename = false;
@@ -222,12 +253,12 @@ split_3 (char *s, size_t s_len,
     ++i;
 
   /* Check for BSD-style checksum line. */
-  algo_name_len = strlen (DIGEST_TYPE_STRING (algorithm));
-  if (strncmp (s + i, DIGEST_TYPE_STRING (algorithm), algo_name_len) == 0)
+  algo_name_len = strlen (DIGEST_TYPE_STRING);
+  if (strncmp (s + i, DIGEST_TYPE_STRING, algo_name_len) == 0)
     {
       if (strncmp (s + i + algo_name_len, " (", 2) == 0)
 	{
-	  *binary = false;
+	  *binary = 0;
 	  return bsd_split_3 (s +      i + algo_name_len + 2,
 			      s_len - (i + algo_name_len + 2),
 			      hex_digest, file_name);
@@ -317,43 +348,46 @@ hex_digits (unsigned char const *s)
 {
   while (*s)
     {
-      if (!ISXDIGIT (*s))
+      if (!isxdigit (*s))
         return false;
       ++s;
     }
   return true;
 }
 
-/* An interface to the function, DIGEST_STREAM,
-   (either md5_stream or sha1_stream).
-   Operate on FILENAME (it may be "-") and put the result in *BIN_RESULT.
+/* An interface to the function, DIGEST_STREAM.
+   Operate on FILENAME (it may be "-").
+
+   *BINARY indicates whether the file is binary.  BINARY < 0 means it
+   depends on whether binary mode makes any difference and the file is
+   a terminal; in that case, clear *BINARY if the file was treated as
+   text because it was a terminal.
+
+   Put the checksum in *BIN_RESULT, which must be properly aligned.
    Return true if successful.  */
 
 static bool
-digest_file (const char *filename, bool binary, unsigned char *bin_result,
-	     int (*digest_stream) (FILE *, void *))
+digest_file (const char *filename, int *binary, unsigned char *bin_result)
 {
   FILE *fp;
   int err;
+  bool is_stdin = STREQ (filename, "-");
 
-  if (STREQ (filename, "-"))
+  if (is_stdin)
     {
       have_read_stdin = true;
       fp = stdin;
-#if O_BINARY
-      /* If we need binary reads from a pipe or redirected stdin, we need
-	 to switch it to BINARY mode here, since stdin is already open.  */
-      if (binary)
-	SET_BINARY (fileno (stdin));
-#endif
+      if (O_BINARY && *binary)
+	{
+	  if (*binary < 0)
+	    *binary = ! isatty (STDIN_FILENO);
+	  if (*binary)
+	    freopen (NULL, "rb", stdin);
+	}
     }
   else
     {
-      /* OPENOPTS is a macro.  It varies with the system.
-	 Some systems distinguish between internal and
-	 external text representations.  */
-
-      fp = fopen (filename, OPENOPTS (binary));
+      fp = fopen (filename, (O_BINARY && *binary ? "rb" : "r"));
       if (fp == NULL)
 	{
 	  error (0, errno, "%s", filename);
@@ -361,7 +395,7 @@ digest_file (const char *filename, bool binary, unsigned char *bin_result,
 	}
     }
 
-  err = (*digest_stream) (fp, bin_result);
+  err = DIGEST_STREAM (fp, bin_result);
   if (err)
     {
       error (0, errno, "%s", filename);
@@ -370,7 +404,7 @@ digest_file (const char *filename, bool binary, unsigned char *bin_result,
       return false;
     }
 
-  if (fp != stdin && fclose (fp) == EOF)
+  if (!is_stdin && fclose (fp) != 0)
     {
       error (0, errno, "%s", filename);
       return false;
@@ -380,18 +414,21 @@ digest_file (const char *filename, bool binary, unsigned char *bin_result,
 }
 
 static bool
-digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
+digest_check (const char *checkfile_name)
 {
   FILE *checkfile_stream;
   uintmax_t n_properly_formatted_lines = 0;
   uintmax_t n_mismatched_checksums = 0;
   uintmax_t n_open_or_read_failures = 0;
-  unsigned char bin_buffer[MAX_DIGEST_BIN_BYTES];
+  unsigned char bin_buffer_unaligned[DIGEST_BIN_BYTES + DIGEST_ALIGN];
+  /* Make sure bin_buffer is properly aligned. */
+  unsigned char *bin_buffer = ptr_align (bin_buffer_unaligned, DIGEST_ALIGN);
   uintmax_t line_number;
   char *line;
   size_t line_chars_allocated;
+  bool is_stdin = STREQ (checkfile_name, "-");
 
-  if (STREQ (checkfile_name, "-"))
+  if (is_stdin)
     {
       have_read_stdin = true;
       checkfile_name = _("standard input");
@@ -407,15 +444,14 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 	}
     }
 
-  SET_MODE (fileno (checkfile_stream), O_TEXT);
   line_number = 0;
   line = NULL;
   line_chars_allocated = 0;
   do
     {
       char *filename;
-      bool binary;
-      unsigned char *hex_digest;
+      int binary;
+      unsigned char *hex_digest IF_LINT (= NULL);
       ssize_t line_length;
 
       ++line_number;
@@ -436,6 +472,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 	line[--line_length] = '\0';
 
       if (! (split_3 (line, line_length, &hex_digest, &binary, &filename)
+	     && ! (is_stdin && STREQ (filename, "-"))
 	     && hex_digits (hex_digest)))
 	{
 	  if (warn)
@@ -444,7 +481,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 		     _("%s: %" PRIuMAX
 		       ": improperly formatted %s checksum line"),
 		     checkfile_name, line_number,
-		     DIGEST_TYPE_STRING (algorithm));
+		     DIGEST_TYPE_STRING);
 	    }
 	}
       else
@@ -457,7 +494,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 
 	  ++n_properly_formatted_lines;
 
-	  ok = digest_file (filename, binary, bin_buffer, digest_stream);
+	  ok = digest_file (filename, &binary, bin_buffer);
 
 	  if (!ok)
 	    {
@@ -476,9 +513,9 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 		 in check file.  Ignore case of hex digits.  */
 	      for (cnt = 0; cnt < digest_bin_bytes; ++cnt)
 		{
-		  if (TOLOWER (hex_digest[2 * cnt])
+		  if (tolower (hex_digest[2 * cnt])
 		      != bin2hex[bin_buffer[cnt] >> 4]
-		      || (TOLOWER (hex_digest[2 * cnt + 1])
+		      || (tolower (hex_digest[2 * cnt + 1])
 			  != (bin2hex[bin_buffer[cnt] & 0xf])))
 		    break;
 		}
@@ -496,8 +533,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
     }
   while (!feof (checkfile_stream) && !ferror (checkfile_stream));
 
-  if (line)
-    free (line);
+  free (line);
 
   if (ferror (checkfile_stream))
     {
@@ -505,7 +541,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
       return false;
     }
 
-  if (checkfile_stream != stdin && fclose (checkfile_stream) == EOF)
+  if (!is_stdin && fclose (checkfile_stream) != 0)
     {
       error (0, errno, "%s", checkfile_name);
       return false;
@@ -515,7 +551,7 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
     {
       /* Warn if no tests are found.  */
       error (0, 0, _("%s: no properly formatted %s checksum lines found"),
-	     checkfile_name, DIGEST_TYPE_STRING (algorithm));
+	     checkfile_name, DIGEST_TYPE_STRING);
     }
   else
     {
@@ -553,20 +589,13 @@ digest_check (const char *checkfile_name, int (*digest_stream) (FILE *, void *))
 int
 main (int argc, char **argv)
 {
-  unsigned char bin_buffer[MAX_DIGEST_BIN_BYTES];
+  unsigned char bin_buffer_unaligned[DIGEST_BIN_BYTES + DIGEST_ALIGN];
+  /* Make sure bin_buffer is properly aligned. */
+  unsigned char *bin_buffer = ptr_align (bin_buffer_unaligned, DIGEST_ALIGN);
   bool do_check = false;
   int opt;
   bool ok = true;
-  bool file_type_specified = false;
-
-#if O_BINARY
-  /* Binary is default on MSDOS, so the actual file contents
-     are used in computation.  */
-  bool binary = true;
-#else
-  /* Text is default of the Plumb/Lankester format.  */
-  bool binary = false;
-#endif
+  int binary = -1;
 
   /* Setting values of global variables.  */
   initialize_main (&argc, &argv);
@@ -581,8 +610,7 @@ main (int argc, char **argv)
     switch (opt)
       {
       case 'b':
-	file_type_specified = true;
-	binary = true;
+	binary = 1;
 	break;
       case 'c':
 	do_check = true;
@@ -592,8 +620,7 @@ main (int argc, char **argv)
 	warn = false;
 	break;
       case 't':
-	file_type_specified = true;
-	binary = false;
+	binary = 0;
 	break;
       case 'w':
 	status_only = false;
@@ -605,10 +632,10 @@ main (int argc, char **argv)
 	usage (EXIT_FAILURE);
       }
 
-  min_digest_line_length = MIN_DIGEST_LINE_LENGTH (algorithm);
-  digest_hex_bytes = DIGEST_HEX_BYTES (algorithm);
+  min_digest_line_length = MIN_DIGEST_LINE_LENGTH;
+  digest_hex_bytes = DIGEST_HEX_BYTES;
 
-  if (file_type_specified & do_check)
+  if (0 <= binary && do_check)
     {
       error (0, 0, _("the --binary and --text options are meaningless when "
 		     "verifying checksums"));
@@ -629,30 +656,23 @@ main (int argc, char **argv)
       usage (EXIT_FAILURE);
     }
 
-  if (do_check)
+  if (!O_BINARY && binary < 0)
+    binary = 0;
+
+  if (optind == argc)
+    argv[argc++] = "-";
+
+  for (; optind < argc; ++optind)
     {
-      if (optind + 1 < argc)
+      char *file = argv[optind];
+
+      if (do_check)
+	ok &= digest_check (file);
+      else
 	{
-	  error (0, 0, _("extra operand %s"), quote (argv[optind + 1]));
-	  fprintf (stderr, "%s\n",
-		   _("Only one operand may be specified when using --check."));
-	  usage (EXIT_FAILURE);
-	}
+	  int file_is_binary = binary;
 
-      ok = digest_check (optind == argc ? "-" : argv[optind],
-			 DIGEST_STREAM (algorithm));
-    }
-  else
-    {
-      if (optind == argc)
-	argv[argc++] = "-";
-
-      for (; optind < argc; ++optind)
-	{
-	  char *file = argv[optind];
-
-	  if (! digest_file (file, binary, bin_buffer,
-			     DIGEST_STREAM (algorithm)))
+	  if (! digest_file (file, &file_is_binary, bin_buffer))
 	    ok = false;
 	  else
 	    {
@@ -667,7 +687,7 @@ main (int argc, char **argv)
 		printf ("%02x", bin_buffer[i]);
 
 	      putchar (' ');
-	      if (binary)
+	      if (file_is_binary)
 		putchar ('*');
 	      else
 		putchar (' ');
