@@ -1,5 +1,5 @@
 /* tail -- output the last part of file(s)
-   Copyright (C) 1989, 90, 91, 1995-2005 Free Software Foundation, Inc.
+   Copyright (C) 1989, 90, 91, 1995-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -51,12 +51,6 @@
 #define AUTHORS \
   "Paul Rubin", "David MacKenzie, Ian Lance Taylor", "Jim Meyering"
 
-#ifndef ENOSYS
-  /* Some systems don't have ENOSYS -- this should be a big enough
-     value that no valid errno value will match it.  */
-# define ENOSYS 99999
-#endif
-
 /* Number of items to tail.  */
 #define DEFAULT_N_LINES 10
 
@@ -203,7 +197,7 @@ static struct option const long_options[] =
   {"lines", required_argument, NULL, 'n'},
   {"max-unchanged-stats", required_argument, NULL, MAX_UNCHANGED_STATS_OPTION},
   {"pid", required_argument, NULL, PID_OPTION},
-  {"presume-input-pipe", no_argument, NULL,
+  {"-presume-input-pipe", no_argument, NULL,
    PRESUME_INPUT_PIPE_OPTION}, /* do not document */
   {"quiet", no_argument, NULL, 'q'},
   {"retry", no_argument, NULL, RETRY_OPTION},
@@ -1021,12 +1015,22 @@ tail_forever (struct File_spec *f, int nfiles, double sleep_interval)
 	      if (old_flags < 0
 		  || (new_flags != old_flags
 		      && fcntl (fd, F_SETFL, new_flags) == -1))
-		error (EXIT_FAILURE, errno,
-		       _("%s: cannot change nonblocking mode"), name);
-	      f[i].blocking = blocking;
+		{
+		  /* Don't update f[i].blocking if fcntl fails.  */
+		  if (S_ISREG (f[i].mode) && errno == EPERM)
+		    {
+		      /* This happens when using tail -f on a file with
+			 the append-only attribute.  */
+		    }
+		  else
+		    error (EXIT_FAILURE, errno,
+			   _("%s: cannot change nonblocking mode"), name);
+		}
+	      else
+		f[i].blocking = blocking;
 	    }
 
-	  if (!blocking)
+	  if (!f[i].blocking)
 	    {
 	      if (fstat (fd, &stats) != 0)
 		{
@@ -1044,7 +1048,7 @@ tail_forever (struct File_spec *f, int nfiles, double sleep_interval)
 		       <= f[i].n_unchanged_stats++)
 		      && follow_mode == Follow_name)
 		    {
-		      recheck (&f[i], blocking);
+		      recheck (&f[i], f[i].blocking);
 		      f[i].n_unchanged_stats = 0;
 		    }
 		  continue;
@@ -1077,7 +1081,8 @@ tail_forever (struct File_spec *f, int nfiles, double sleep_interval)
 	    }
 
 	  bytes_read = dump_remainder (name, fd,
-				       blocking ? COPY_A_BUFFER : COPY_TO_EOF);
+				       (f[i].blocking
+					? COPY_A_BUFFER : COPY_TO_EOF));
 	  any_input |= (bytes_read != 0);
 	  f[i].size += bytes_read;
 	}
@@ -1450,7 +1455,8 @@ parse_options (int argc, char **argv,
 {
   int c;
 
-  while ((c = getopt_long (argc, argv, "c:n:fFqs:v", long_options, NULL))
+  while ((c = getopt_long (argc, argv, "c:n:fFqs:v0123456789",
+			   long_options, NULL))
 	 != -1)
     {
       switch (c)
@@ -1548,13 +1554,18 @@ parse_options (int argc, char **argv,
 
 	case_GETOPT_VERSION_CHAR (PROGRAM_NAME, AUTHORS);
 
+	case '0': case '1': case '2': case '3': case '4':
+	case '5': case '6': case '7': case '8': case '9':
+	  error (EXIT_FAILURE, 0,
+		 _("option used in invalid context -- %c"), c);
+
 	default:
 	  usage (EXIT_FAILURE);
 	}
     }
 
   if (reopen_inaccessible_files && follow_mode != Follow_name)
-    error (0, 0, _("warning: --retry is useful only when following by name"));
+    error (0, 0, _("warning: --retry is useful mainly when following by name"));
 
   if (pid && !forever)
     error (0, 0,
@@ -1624,12 +1635,19 @@ main (int argc, char **argv)
       file = &dummy_stdin;
 
       /* POSIX says that -f is ignored if no file operand is specified
-	 and standard input is a pipe.  */
-      if (forever)
+	 and standard input is a pipe.  However, the GNU coding
+	 standards say that program behavior should not depend on
+	 device type, because device independence is an important
+	 principle of the system's design.
+
+	 Follow the POSIX requirement only if POSIXLY_CORRECT is set.
+	 Ideally this would ignore -f only for pipes, but S_ISFIFO
+	 succeeds for both FIFOs and pipes and we know of no portable,
+	 reliable way to distinguish them.  */
+      if (forever && getenv ("POSIXLY_CORRECT"))
 	{
 	  struct stat stats;
-	  if (fstat (STDIN_FILENO, &stats) == 0
-	      && IS_PIPE_LIKE_FILE_TYPE (stats.st_mode))
+	  if (fstat (STDIN_FILENO, &stats) == 0 && S_ISFIFO (stats.st_mode))
 	    forever = false;
 	}
     }
